import logging
from pathlib import Path

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

# Read lime configuration .toml
_inst_dir = Path(__file__).parent
_conf_path = _inst_dir/'config.toml'
with open(_conf_path, mode="rb") as fp:
    _setup_cfg = tomllib.load(fp)

from .io import save_dataset

from .main import DataSet, Grid


import logging
from pathlib import Path
from . import _setup_cfg

_logger = logging.getLogger('Innate')

REF_entry = _setup_cfg['parameter_labels']['param']

try:
    from astropy.io import fits
    astropy_check = True
except ImportError:
    astropy_check = False

try:
    import h5netcdf
    h5netcdf_check = True
except ImportError:
    h5netcdf_check = False


# Define library error function
class InnateError(Exception):
    """InnateError exception function"""


def load_dataset(fname: str):

    # Check the file location
    fname = Path(fname)
    ext = fname.suffix

    if not fname.is_file():
        raise InnateError(f'Not dataset file found at {fname}')

    # Run the appropriate file saving workflow
    match ext:

        case '.fits':
            return fits_file_load(fname)

        case '.nc':
            return h5netcdf_file_load(fname)

        case _:
            return KeyError(f'The extension "{ext}" is not recognized, please use ".nc" or ".fits"')


def save_dataset(fname: str, grid_dict: dict, common_cfg: dict, custom_cfg: dict):

    """

    :param fname:
    :type str or Pathlib: path.Pathlib

    :param cfg_dict:

    :param grid_dict:

    """

    # Check the file location
    fname = Path(fname)
    ext = fname.suffix

    # Check there is input data
    if len(grid_dict) == 0:
        raise InnateError(f'Output data grid does not contain any entry')

    # Run the appropriate file saving workflow
    match ext:

        case '.fits':
            return fits_file_save(fname, grid_dict, common_cfg, custom_cfg)

        case '.nc':
            return h5netcdf_file_save(fname, grid_dict, common_cfg, custom_cfg)

        case _:
            return KeyError(f'The extension "{ext}" is not recognized, please use ".nc" or ".fits"')


def fits_file_save(fname: str, grid_dict: dict, common_cfg: dict, custom_cfg: dict):

    if astropy_check:

        # Create a primary HDU
        hdu_list = fits.HDUList([fits.PrimaryHDU()])

        # Generate the fits file
        for key, grid in grid_dict.items():
            hdu_i = fits.ImageHDU(grid, name=key)
            hdu_list.append(hdu_i)

        # Write the fits file
        hdu_list.writeto(fname, overwrite=True)

    else:
        raise InnateError(f'To open ".fits" files (astropy.io.fits) you need to install the astropy package')

    return


def fits_file_load(fname: str):

    grid_dict, cfg_dict = {}, {}

    if astropy_check:
        with fits.open(fname) as hdu_list:
            for i in range(1, len(hdu_list)):
                grid_dict[hdu_list[i].name] = hdu_list[i].data
    else:
        raise InnateError(f'To open ".fits" files (astropy.io.fits) you need to install the astropy package')

    return grid_dict, cfg_dict


def h5netcdf_file_save(fname: str, grid_dict: dict, common_cfg: dict, custom_cfg: dict):

    if h5netcdf_check:

        # Use the first item for the dimensions # TODO this might cause issues if grids have different size
        grid_0 = grid_dict[list(grid_dict.keys())[0]]
        data_shape = grid_0.shape

        if len(data_shape) != len(common_cfg['axes']):
            raise InnateError(f'Data set shape ({data_shape}) size is different from configuration axes'
                              f' ({common_cfg["axes"]}) size')

        with h5netcdf.File(fname, 'w') as f:

            # Unique dimensions for all the datase
            for i, axis in enumerate(common_cfg['axes']):
                f.dimensions[axis] = data_shape[i]

            # Iterate over the data grids
            for grid_name, grid in grid_dict.items():
                var = f.create_variable(grid_name, common_cfg['axes'], data=grid)
                var.attrs[REF_entry] = str(grid_name)

                # Common grid entries
                if common_cfg is not None:
                    for key, value in common_cfg.items():
                        var.attrs[key] = value

                # Local entries
                if custom_cfg is not None:
                    local_cfg = custom_cfg.get(grid_name)
                    if local_cfg is not None:
                        for key, value in local_cfg.items():
                            var.attrs[f'{grid_name}_{key}'] = value

    else:
        raise InnateError(f'To open ".nc" (h5netcdf) files you need to install the h5netcdf package')

    return


def h5netcdf_file_load(fname: str):

    grid_dict, common_cfg, local_cfg = {}, {}, {}

    if h5netcdf_check:
        with h5netcdf.File(fname, 'r') as f:

            # Loop through the grid
            for var_name in f.variables:

                # Unpack the library container
                var = f.variables[var_name]

                # Scientific data
                grid_dict[var_name] = var[...]

                # Unpack the configuration parameters
                common_cfg[var_name], local_cfg[var_name] = {}, {}
                for attr in var.attrs:

                    # Global conf
                    if not attr.startswith(var_name):
                        common_cfg[var_name][attr] = var.attrs[attr]

                    # Local conf
                    else:
                        local_cfg[var_name][attr[len(var_name)+1:]] = var.attrs[attr]

                # Convert to None if empty configuration
                if len(grid_dict[var_name]) == 0:
                    grid_dict[var_name] = None

                if len(common_cfg[var_name]) == 0:
                    common_cfg[var_name] = None

                if len(local_cfg[var_name]) == 0:
                    local_cfg[var_name] = None

    else:
        raise InnateError(f'To open ".nc" (h5netcdf) files you need to install the h5netcdf package')

    # Parse the file
    return grid_dict, common_cfg, local_cfg






import logging
import re
import numpy as np
from .. import _setup_cfg

_logger = logging.getLogger('Innate')


def generate_specific_function(expression, coefficients, variable_names):
    # Define the function to only vary with specified variables
    def specific_function(*variable_values):
        local_vars = coefficients.copy()  # Copy the dictionary of fixed variables
        # Update variable names with values dynamically from the input
        local_vars.update(dict(zip(variable_names, variable_values)))
        local_vars['np'] = np  # Ensure np is available for np.log10 and other operations
        return eval(expression, {}, local_vars)
    # The number of inputs is now the number of variable names provided
    return np.frompyfunc(specific_function, len(variable_names), 1)


def extract_coef_names(expression):
    # Regex to find isolated letters (considered variables)
    pattern = r'\b[a-zA-Z]\b'
    # Find all matches and return as a set to remove duplicates
    matches = set(re.findall(pattern, expression))
    return sorted(matches)


def create_coef_dict (coef_names, coef_values):
    if len(coef_names)==len(coef_values):
       out = dict(zip(coef_names, coef_values))
    else:
        raise TypeError("length of coefficietns names different from the length of coefficients values")
    return(out)


def extract_variables_names(expression, suffix='_range'):
    # Construct the regex pattern dynamically to find words ending with the given suffix
    pattern = rf'\b\w+{re.escape(suffix)}\b'  # Use re.escape to safely include the suffix in the regex
    # Find all matches and return as a set to remove duplicates
    matches = set(re.findall(pattern, expression))
    return sorted(list(matches))  # Return a sorted list of unique matches


def parse_string_equation(data_label, str_eqn, coeffs_eqn, variable_names):

    if (str_eqn is not None) or (coeffs_eqn is not None):
        coef_names = extract_coef_names(str_eqn)
        coeffs_dict = create_coef_dict(coef_names, coeffs_eqn)
        eqn = generate_specific_function(str_eqn, coeffs_dict, variable_names)

    else:
        message = f'Data set "{data_label}" is missing:'
        if str_eqn is None:
            message += f'\nParametrisation formula ("eqn" key in dataset configuration).'
        if coeffs_eqn is None:
            message += f'\nParametrisation coefficients ("eqn_coeffs" key in dataset configuration).'
        _logger.warning(message)

        eqn, coeffs = None, None

    return eqn, coeffs_dict


class Regressor:

    def __init__(self, grid, technique_list, data_cfg=None):

        self.eqn = None
        self.coeffs = None
        self.techniques = []

        # Constrain to regresion techniques
        algorithms = list(set(_setup_cfg['parameter_labels']['reg'].keys()) & set(technique_list))

        # Regular grid Interpolation
        if 'eqn' in algorithms:
            self.techniques.append('eqn')

            # Reconstruct the string equation into a python function
            self.eqn, self.coeffs = parse_string_equation(grid.label,
                                                          data_cfg.get('eqn', None),
                                                          data_cfg.get('eqn_coeffs', None),
                                                          data_cfg.get('axes', None))

        return







import logging
import numpy as np
from pathlib import Path
from .io import InnateError, load_dataset
from .approximation import Approximator

_logger = logging.getLogger('Innate')

def reconstruct_axes_range(data_label, axes, data_cfg, data_shape):

    axes_range = {}
    for i, dim in enumerate(axes):

        linspace_idcs = data_cfg.get(f'{dim}_range')
        if linspace_idcs is None:
            raise InnateError(
                f'The input grid "{data_label}" configuration does not include a range for the axis "{dim}"')

        axes_range[dim] = np.linspace(linspace_idcs[0], linspace_idcs[1], linspace_idcs[2])

        # Check the sizes are the same
        if axes_range[dim].size != data_shape[i]:
            _logger.warning(f'The parameter "{dim}" input dimensions ({data_shape[i]}) and '
                            f'axes dimensions ({axes_range[dim].size}) are different ')

    return axes_range


class Grid:

    def __init__(self, grid_label, data_array, data_cfg, tensor_library='pytensor'):

        # Class attributes
        self.label = None
        self.description = None
        self.data = None
        self.axes = None
        self.shape = None
        self.axes_range = None

        # Assign attribute values
        self.label = grid_label
        self.description = data_cfg['parameter']
        self.data = data_array
        self.axes = data_cfg['axes']
        self.shape = self.data.shape
        self.axes_range = reconstruct_axes_range(grid_label, self.axes, data_cfg, self.shape)

        # Declare the function attributes treatments
        approx_techniques = data_cfg.get('approximation')
        self.approx = Approximator(self, approx_techniques, data_cfg)

        # # Initiate data and interpolators
        # grid_path = Path(grid)
        # self.grid = load_dataset(grid_path) if grid_path.is_file() else grid

        # if interpolator is not None:

            # print(f'\n- Compiling {interpolator} interpolator')
            #
            # if interpolator == 'pytensor':
            #     self.lib_interpl = interpolator
            #     self.x_range = np.linspace(x_space[0], x_space[1], x_space[2])
            #     self.y_range = np.linspace(y_space[0], y_space[1], y_space[2])
            #     self.interpl = interpolation_selection(self.grid, self.x_range, self.y_range, z_range=None,
            #                                            interp_type='point')
            #     print('-- done')

        return


class DataSet(dict):

    def __init__(self, array_dict, common_cfg, local_cfg, **kwargs):

        # Attributes
        self.data_labels = None
        self.shape_array = None

        # Unpack the individual grids into the class dictionary
        self._compile_grids(array_dict, common_cfg, local_cfg, **kwargs)

        return

    @classmethod
    def from_file(cls, fname, grid_cfg=None):

        # Load and parse the input data
        array_dict, common_cfg, local_cfg = load_dataset(fname)

        # Update the input configuration with the parameters from the user

        return cls(array_dict, common_cfg, local_cfg)

    def _compile_grids(self, array_dict, common_cfg, local_cfg, **kwargs):

        # Extract the keys from the data and configuration containers to create a master list for the Grid set
        common_set = set(array_dict.keys()) & set(common_cfg.keys()) & set(local_cfg.keys())
        self.data_labels = np.array(list(common_set))

        # Loop throught the dataset and create the grids
        for i, data_label in enumerate(self.data_labels):

            # Extract the Grid data
            grid_array = array_dict.get(data_label)
            data_glob_conf = common_cfg.get(data_label)
            data_local_conf = local_cfg.get(data_label)

            # Local entries update the new
            grid_cfg = data_glob_conf if data_local_conf is None else {**data_glob_conf, **data_local_conf}
            self[data_label] = Grid(data_label, grid_array, grid_cfg, **kwargs)

        return

    def extract_approximation(self, technique, label_list=None):

        label_list = label_list if label_list is not None else list(self.keys())

        approx_dict = {}
        for label in label_list:

            if technique in self[label].approx.interp.techniques:
                approx_dict[label] = getattr(self[label].approx.interp, technique)

            elif technique in self[label].approx.reg.techniques:
                approx_dict[label] = getattr(self[label].approx.reg, technique)

            else:
                _logger.critical(f'Input approximation "{technique}" is not available for dataset "{label}"')

        return approx_dict




import logging
import numpy as np
from matplotlib import pyplot as plt, rc_context
from . import _setup_cfg

_logger = logging.getLogger('Innate')


class Themer:

    def __init__(self, conf, style='default', library='matplotlib'):

        # Attributes
        self.conf = None
        self.style = None
        self.base_conf = None
        self.colors = None
        self.library = None

        # Assign default
        self.conf = conf.copy()
        self.library = library
        self.set_style(style)

        return

    def fig_defaults(self, user_fig=None, fig_type=None):

        # Get plot configuration
        if fig_type is None:
            fig_conf = self.base_conf.copy()
        else:
            fig_conf = {** self.base_conf, **self.conf[self.library][fig_type]}

        # Get user configuration
        fig_conf = fig_conf if user_fig is None else {**fig_conf, **user_fig}

        return fig_conf

    def ax_defaults(self, user_ax, x_units, y_units, norm_flux, fig_type='default', **kwargs):

        # Default wavelength and flux
        if fig_type == 'plane':

            # Spectrum labels x-wavelegth, y-flux # TODO without units
            ax_cfg = {'xlabel': f'{x_units:s}', 'ylabel': f'{y_units:s}'}

            # Update with the user configuration
            ax_cfg = ax_cfg if user_ax is None else {**ax_cfg, **user_ax}

        # Spatial cubes
        elif fig_type == 'cube':

            ax_cfg = {} if user_ax is None else user_ax.copy()

            # Define the title
            if ax_cfg.get('title') is None:

                title = r'{} band'.format(kwargs['line_bg'].latex_label[0])

                line_fg = kwargs.get('line_fg')
                if line_fg is not None:
                    title = f'{title} with {line_fg.latex_label[0]} contours'

                if len(kwargs['masks_dict']) > 0:
                    title += f'\n and spatial masks at foreground'

                ax_cfg['title'] = title

            # Define x axis
            if ax_cfg.get('xlabel') is None:
                ax_cfg['xlabel'] = 'x' if kwargs['wcs'] is None else 'RA'

            # Define y axis
            if ax_cfg.get('ylabel') is None:
                ax_cfg['ylabel'] = 'y' if kwargs['wcs'] is None else 'DEC'

            # Update with the user configuration
            ax_cfg = ax_cfg if user_ax is None else {**ax_cfg, **user_ax}

        # No labels
        else:
            ax_cfg = {}

            # Update with the user configuration
            ax_cfg = ax_cfg if user_ax is None else {**ax_cfg, **user_ax}

        return ax_cfg

    def set_style(self, style=None, fig_cfg=None, colors_conf=None):

        # Set the new style
        if style is not None:
            self.style = np.atleast_1d(style)
        else:
            self.style = np.atleast_1d('default')

        # Generate the default
        self.base_conf = self.conf[self.library]['default'].copy()
        for style in self.style:
            self.base_conf = {**self.base_conf, **self.conf[self.library][style]}

        # Add the new configuration
        if fig_cfg is not None:
            self.base_conf = {**self.base_conf, **fig_cfg}

        # Set the colors
        for i_style in self.style:
            if i_style in self.conf['colors']:
                self.colors = self.conf['colors'][style].copy()

        # Add the new colors
        if colors_conf is not None:
            self.colors = {**self.colors, **colors_conf}

        if self.colors is None:
            _logger.warning(f'The input style {self.style} does not have a LiMe color database')

        return


# LiMe figure labels and color formatter
theme = Themer(_setup_cfg)

